@@ -4,7 +4,7 @@
 
 #include "config2.h"
 #include "common.h"
-#include "core.h"
+#include "sctcore.h"
 
 #include <stdio.h>
 #include <stdlib.h>
@@ -72,7 +72,7 @@ pid_t pidchild;
 pthread_mutex_t pidmutex, tkill_mx, tdisplay_mx, judge_tle_mx;
 pthread_t tkill, tdisplay;
 char tkill_yes = 0, tdisplay_yes = 0, judge_tle = 0;
-sem_t mcthr, addthr;
+sem_t mcthr, tlethr, dispthr;
 
 static const char* childmsg_text[SCTCHILD_MSGMAX] = {
 	"開啟輸入檔案",
@@ -99,7 +99,7 @@ static void sctjudge_makechild_cleanup_p1(void){
 	if(tkill_yes){
 		tkill_yes = 0;
 		pthread_mutex_unlock(&tkill_mx);
-		pthread_cancel(tkill);
+		sem_post(&tlethr);
 	}else{
 		pthread_mutex_unlock(&tkill_mx);
 	}
@@ -108,10 +108,11 @@ static void sctjudge_makechild_cleanup_p1(void){
 	if(tdisplay_yes){
 		tdisplay_yes = 0;
 		pthread_mutex_unlock(&tdisplay_mx);
-		pthread_cancel(tdisplay);
+		sem_post(&dispthr);
 	}else{
 		pthread_mutex_unlock(&tdisplay_mx);
 	}
+
 }
 	
 static void sctjudge_makechild_cleanup_p2(mcopt, oldperm, copiedexe)
@@ -227,23 +228,23 @@ static void sctjudge_list_settings(const struct makechildopt* mcopt){
 	printf("\t執行時間限制：%d 毫秒\n", mcopt->exectime);
 	fputs("\t記憶體使用量限制：", stdout);
 	if(mcopt->memlimit > 0){
-		printf("%d MiB\n", mcopt->memlimit);
+		printf("%d 位元組\n", mcopt->memlimit);
 	}else{
 		puts("無限制");
 	}
 	fputs("\t輸出限制：", stdout);
 	if(mcopt->outlimit > 0){
-		printf("%d MiB\n", mcopt->outlimit);
+		printf("%d 位元組\n", mcopt->outlimit);
 	}else{
 		puts("無限制");
 	}
 	if(mcopt->flags & SCTMC_SETUID){
-		printf("\t設定受測程式執行時的 UID 為 %d\n", mcopt->uid);
+		printf("\t設定受測程式執行時的 UID 為 %u\n", mcopt->uid);
 	}else{
 		puts("\t執行受測程式時維持原有的 real UID 和 effective UID");
 	}
 	if(mcopt->flags & SCTMC_SETGID){
-		printf("\t設定受測程式執行時的 GID 為 %d\n", mcopt->gid);
+		printf("\t設定受測程式執行時的 GID 為 %u\n", mcopt->gid);
 	}else{
 		puts("\t執行受測程式時維持原有的 real GID、effective GID "
 				"和 supplementary GID");
@@ -273,7 +274,6 @@ static int read_size(int fd, void* buf, size_t count){
 void* sctjudge_makechild(void* arg){
 	/* 首先等所有 thread 都啟動完成再開始動作 */
 	sem_wait(&mcthr);
-	sem_destroy(&mcthr);
 
 	/* 宣告變數囉！ */
 	int i;
@@ -377,7 +377,11 @@ void* sctjudge_makechild(void* arg){
 	/* 建立用來和 chlid process 溝通的 pipe */
 	pipe(childpipe);
 	pthread_mutex_lock(&pidmutex);
+#ifdef HAVE_WORKING_FORK
 	pidchild = fork();
+#else
+#error "This program requires a working fork function."
+#endif
 	if(pidchild < 0){
 #ifndef HAVE_CONF_CAP
 		disable_setuid();
@@ -421,10 +425,9 @@ void* sctjudge_makechild(void* arg){
 		close(childpipe[0]);
 
 		/* 現在我們確定受測程式已經在執行了，所以需要記錄一下時間
-		 * 通知其他 thread，然後就可以把 semaphore 丟掉了 */
-		sem_post(&addthr);
-		sem_post(&addthr);
-		sem_destroy(&addthr);
+		 * 通知其他 thread */
+		sem_post(&tlethr);
+		sem_post(&dispthr);
 
 		/* 開始要 wait 了，莫名其妙 stop 的程式我最多只會送
 		 * 5 次 SIGCONT 給你（避免進入無限迴圈） 
@@ -535,7 +538,7 @@ void* sctjudge_makechild(void* arg){
 #ifndef HAVE_CONF_CAP
 		enable_setuid();
 #endif
-		fchown(fdoutput, getuid(), -1);
+		fchown(fdoutput, procrealuid, -1);
 
 		/* 再來就是 stderr 了 */
 		if(mcopt.flags & SCTMC_REDIR_STDERR){
@@ -546,7 +549,7 @@ void* sctjudge_makechild(void* arg){
 				SCTCHILD_WRITE_SUCCMSG(SCTCHILD_DUP2_STDERR);
 			}
 		}else{
-			fderr = open(NULL_DEVICE, O_RDONLY);
+			fderr = open(WITH_NULL, O_RDONLY);
 			if(fderr < 0){
 				SCTCHILD_WRITE_FAILMSG(SCTCHILD_OPEN_STDERR);
 				exit(1);
@@ -615,13 +618,13 @@ void* sctjudge_makechild(void* arg){
 #ifndef HAVE_CONF_CAP
 		else{
 			/* 確保 setuid 可執行檔所造成的 effective UID 改變不會傳給子程序 */
-			setuid(getuid());
+			setuid(procrealuid);
 		}
 #endif
 
 		/* 開始設定資源限制 */
 		if(mcopt.memlimit > 0){
-			childressize = mcopt.memlimit * 1024 * 1024;
+			childressize = mcopt.memlimit;
 			childres.rlim_cur = childressize;
 			childres.rlim_max = childressize;
 			if(setrlimit(RLIMIT_AS, &childres) < 0){
@@ -633,7 +636,7 @@ void* sctjudge_makechild(void* arg){
 		}
 
 		if(mcopt.outlimit > 0){
-			childressize = mcopt.outlimit * 1024 * 1024;
+			childressize = mcopt.outlimit;
 			childres.rlim_cur = childressize;
 			childres.rlim_max = childressize;
 			if(setrlimit(RLIMIT_FSIZE, &childres) < 0){
